import * as React from "react";

import submitMSISDN, { IConfig, getRedirectUrl, ILinkExtraParams } from "./main";
import * as RDS from "../../common-types/RemoteDataState";
import { ITracker } from "../../pacman/record";
import { IFlow } from "../super-strategy/HOC";

export const Link: React.ComponentType<{ href: string, children: React.ReactNode } & React.HTMLAttributes<HTMLAnchorElement>> = ({ href, children, ...props }) =>
  <a href={href} {...props}>{children}</a>

export type MSISDNEntryFailure = {
  errorType: MSISDNEntryErrorTypes;
  error: any;
};
export type MSISDNEntrySuccess = { Link: ACompType, href: string }
type ACompType =
  React.ComponentType<{ children: React.ReactNode } & React.HTMLAttributes<HTMLAnchorElement>>


export type IdentifyUserErrorTypes = "NotOn3g" | "UnknownError";
export type SubscribeUserErrorTypes = "SubscriptionError" | "UnknownError";
  
export type IdentifyFailure = {
  errorType: IdentifyUserErrorTypes;
  error: Error;
};


export type IdentifySuccess = { Link: ACompType, href: string }

export type SubscribeSuccess = {
  product_url:string;
};

export type SubscribeFailure = {
  errorType: SubscribeUserErrorTypes;
  error: Error;
};


export type IState =
| {
    type: "IDENTIFY";
    result: RDS.RemoteDataState<IdentifyFailure, IdentifySuccess>;
  }
| {
  type: "MSISDNEntry";
  result: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }>;
}
  
export type MSISDNEntryErrorTypes =
  | "AlreadySubscribed"
  | "UnknownError"
  | "InvalidMSISDN";


// const mkLink = (tracker: ITracker, href: string) => ({ children, keyword, shortcode, ...props }) => {
//   return <Link onClick={() => {
//     if (!tracker) {
//       console.warn("tracker parameter is null in mkLink")
//     } else {
//       tracker.advancedInFlow("click2sms", "click", href)
//     }
//   }}
//     href={href} {...props}
//   >{children}</Link>
// }

const clickLinkComp = (tracker: ITracker, url: string) => ({children, ...props}) => <a {...props} onClick={(...args) => {
  tracker.advancedInFlow('tallyman.v1-click-redir', 'click', {url})
  !!props.onClick ? props.onClick(...args) : void 6;
}} href={url}>{children}</a>

export type PINEntryErrorTypes = "UnknownError" | "TooEarly" | "InvalidPIN";

export const initialState: IState = { type: "MSISDNEntry", result: RDS.NothingYet<MSISDNEntryFailure, MSISDNEntrySuccess>() }
export const mockedMSISDNEntrySuccess: IState = { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess>({ href: 'https://www.google.com', Link: clickLinkComp(null, 'https://www.google.com') }) }
export const mockedMSISDNEntryFailure: IState = { type: "MSISDNEntry", result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({ errorType: "InvalidMSISDN", error: "Invalid Mobile Number" }) }
// export const mockedCompletedState: IState = { type: "Completed", result: void 6 }

// export const mockedCompletedState: IState = { type: "Completed", result: void 6 }


export function matchMSISDNEntry<R>(m: RDS.IMatcher<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }, R>) { return RDS.match<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }, R>(m) }

export interface IActions {
  submitMSISDN: (window: Window, config: IConfig, internationalMsisdn: string, bupperMsisdn: string, strategy?:IFlow) => void,
  backToStart: () => void,
  identifyAction:()=>any,
}

export type HOCProps = {
  currentState: IState;
  actions: IActions;
  strategyConfig?:IFlow;
  ClickLink: ACompType;
};


export function match<R>(
  { identify, msisdnEntry }: 
  {   identify: (rds: RDS.RemoteDataState<IdentifyFailure, IdentifySuccess>) => R
    , msisdnEntry: (rds: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess>) => R
  }): (state: IState) => R {
  return state => {
    switch (state.type) {
      case 'IDENTIFY':
        return identify(state.result)
      case 'MSISDNEntry':
        return msisdnEntry(state.result)
    }
  }
}

export default <P extends HOCProps>(tracker: ITracker, maybeConfig: IConfig, Comp: React.ComponentType<P>, extraParams?:ILinkExtraParams, strategy?:IFlow) => (
  initialState: IState
  , isMocked?:boolean
) => {
  const HOC = class HOC extends React.PureComponent<P> {
    state: {
      currentState: IState;
      actions: IActions;
      redirectUrl:string;
    };
    constructor(props) {
      super(props);
      const self = this;
      this.state = {
        currentState: initialState,
        actions: {
          backToStart: () => {
            this.setState({
              currentState: initialState
            })
          },
          submitMSISDN: async (...args) => {
            this.setState({
              currentState: { type: "MSISDNEntry", result: RDS.Loading() } as IState
            })
            const [, , msisdn] = args
            tracker.advancedInFlow('tallyman.v1-click-redir', 'msisdn-submitted', { msisdn })
            if(strategy){
              args[4] = strategy;
            }

            try {
              const href = await submitMSISDN(...args);
              tracker.advancedInFlow('tallyman.v1-click-redir', 'msisdn-submission-success', { msisdn })

              const clickLink = clickLinkComp(tracker, href)

              self.setState({
                currentState: { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }>({ href, Link: clickLink }) } as IState,
                actions: { ...self.state.actions }
              })
            } catch (ex) {
              const errorType: MSISDNEntryErrorTypes =
                "SEAlreadySubscribed" === ex.type
                  ? "AlreadySubscribed"
                  : "SEInvalidMSISDN" == ex.type
                    ? "InvalidMSISDN"
                    : "UnknownError";
              self.setState({
                currentState: {
                  type: "MSISDNEntry",
                  result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({
                    errorType: errorType,
                    error: ex
                  })
                } as IState
              });
              tracker.recedeInFlow('tallyman.v1-click-redir', 'msisdn-submission-failure', { msisdn, errorType: errorType || 'UnknownError' })
            }
          },
          identifyAction: async ()=>{
            try{
              self.setState({
                currentState:{ type:"IDENTIFY", result: RDS.Loading()} as IState
              });
              const href = await getRedirectUrl(maybeConfig || {}, extraParams, strategy);


              const clickLink = clickLinkComp(tracker, href)

              self.setState({
                currentState: { type: "IDENTIFY", result: RDS.Success<IdentifyFailure, IdentifySuccess & { Link: ACompType }>({ href, Link: clickLink }) } as IState,
              });
              tracker.customEvent('Flow', 'advance-auto', 'identify-success', {redirect_url:href})
            }catch(ex) {
              console.error(ex)
              const errorType: IdentifyUserErrorTypes =
                "NotOn3g" === ex.type
                  ? "NotOn3g"
                  : "UnknownError";
              
              tracker.customEvent('Flow', 'recede-auto', 'identify-failure', {error:ex.toString(), errorType})
              this.setState({
                currentState:{
                  type: "MSISDNEntry",
                  result: RDS.NothingYet<MSISDNEntryFailure, MSISDNEntrySuccess>()
                } as IState
              });
            }

          }
        },
        redirectUrl:""
      };
    }

    async UNSAFE_componentWillMount() {
      {
        !isMocked &&
        await this.state.actions.identifyAction()
      }
    }
    render() {
      return (
        <Comp
          {...this.props}
          currentState={this.state.currentState}
          actions={this.state.actions}
          strategyConfig={strategy ? strategy : undefined}
        />
      );
    }
  };
  HOC.prototype['_client'] = "tallyman.v1-click-redir";

  return HOC;
}