import { IFlow } from "../super-strategy/HOC";


export type IKeywordShortcode = {
  keyword : string,
  shortcode : string,
  operator?: string
}

type Config = {
  host?: string;
  country?: string;
  slug?: string;
  automaticallySubmitAllOperators?: boolean;
  operators?: string[];
}
export type IConfig = 
  | { tag: "tallyman"} & Config 
  | { tag: "keywordAndShortCode" } & IKeywordShortcode
  | { tag: "keyword" } & Config


export type SingleSubmissionResult = IKeywordShortcode & {
  type: "SingleSubmissionResult"
}

export type MultiOperatorSubmissionResult =  {
  type: "MultiOperatorSubmissionResult",
  data: Array<IKeywordShortcode>
}

type ICheckSubscriptionResult = {
  success: false;
  rockman_id: string;
  message: string;
} | {
  success: true;
  msisdn: string;
  product_url?: string;
  rockman_id: string;
}
const getConfig = ():Config=>{
  try {
    const configs:Config = require("../../../../../config.json");
    return configs;
  } catch (ex) {
    throw "config.json not found";
  }
};


type IFetchResult = { "success": boolean, "uid": string, "shortcode": string, "keyword": string, displayKeyword:string, message?: string }


async function loadOnce(window: Window, maybeConfig: IConfig, operator?:string, keyword?: string, shortcode?: string, strategy?:IFlow): Promise<SingleSubmissionResult> {
  
  if(!!maybeConfig && maybeConfig.tag == "keywordAndShortCode") {
    const {keyword, shortcode} = maybeConfig
    return {
      type: "SingleSubmissionResult",
      ...{keyword, shortcode} as IKeywordShortcode
    }
  } else {
    const offer = window.pac_analytics.visitor.offer
    const config = strategy ? {
      offer:window.pac_analytics.visitor.offer,
      ...strategy.flowConfig
    } : {...getConfig(), ...(maybeConfig || {})}
    const { host, country, slug } = config
    const rockmanId = window.pac_analytics.visitor.rockmanId
    const search = window.location.search.substr(1) || ''
    const result: IFetchResult = await fetch(`https://${host}/tallyman/v1/?action=oc2sms&country=${country}&slug=${slug}&offerId=${offer}${!!keyword ? '&keyword=' + keyword : ''}&device=smart&rockman_id=${rockmanId}${!!operator ? '&operator=' + operator : ''}&${search}`).then(x => x.json())
    if(!result.success) {
      throw result.message;
    } else {
      return {
        type:"SingleSubmissionResult",
        ...{keyword: `${keyword || result.displayKeyword}${!!result.uid ? ' ' + result.uid : ''}`, shortcode: result.shortcode}
      }
      
    }
  }
  
}
type Either<L, R> = { type: 'left', value: L } | {type: 'right', value: R}
function right<L, R>(r: R) {
  return {type: 'right', value: r} as Either<L, R>
}
function left<L, R>(l: L) {
  return { type: 'left', value: l } as Either<L, R>
}
function foldEither<L, R, T>(left: (l:L) => T, right: (r:R) => T, either: Either<L, R>) : T {
  return either.type == 'left' 
    ? left(either.value)
    : right(either.value)
}
async function loadForMultipleOperators(window: Window, maybeConfig: IConfig, operators: string[], keyword?: string, shortcode?:string, strategy?:IFlow): Promise<MultiOperatorSubmissionResult> {

    const results = Promise.all(operators.map(async o => {
      try {
        const singleResult = await loadOnce(window, maybeConfig, o, keyword, shortcode, strategy)
        //console.log("singleResult", singleResult)
        return right({...singleResult, operator: o})
      } catch (error) {
        return left(error)
      }
    }))

    const finalResults = await Promise.resolve(results);
    if (finalResults.every(r => r.type == 'left')) {
      // no submission worked
      throw finalResults[0].value;
    }
    else{
      const arrResult = [];
      finalResults.map((obj)=>{
        if(obj.type == 'right'){
          const { keyword, shortcode, operator } = obj.value as IKeywordShortcode
          arrResult.push({keyword, shortcode, operator })
        }
      });

      return {
        type:"MultiOperatorSubmissionResult",
        data: arrResult
      } 
    }
}


export default async function load(window: Window, maybeConfig: IConfig, keyword?: string, shortcode?: string, strategy?:IFlow): Promise<SingleSubmissionResult | MultiOperatorSubmissionResult> {
  if (!!maybeConfig && maybeConfig.tag == "keywordAndShortCode") {
    return {
      type: "SingleSubmissionResult",
      ...{ keyword:maybeConfig.keyword, shortcode:maybeConfig.shortcode } as IKeywordShortcode
    }
  } else if (!!maybeConfig && maybeConfig.tag == "keyword") {
    return load1(window, maybeConfig, keyword, shortcode, strategy)
  } else {
    return load1(window, maybeConfig, null, null, strategy)
  }
}

export async function load1(window: Window, maybeConfig: IConfig, keyword?: string, shortcode?: string, strategy?:IFlow): Promise<SingleSubmissionResult | MultiOperatorSubmissionResult> {  
  const config: Config = strategy ? {
    ...strategy.flowConfig
  } : {...getConfig(), ...(maybeConfig || {})}
  if(true == config.automaticallySubmitAllOperators && !!config.operators && config.operators.length > 0) {
    return loadForMultipleOperators(window, maybeConfig, config.operators, keyword, shortcode, strategy)
  } else {
    return loadOnce(window, maybeConfig, null, keyword, shortcode, strategy)
  }
}

const wait = (ms: number) =>
  new Promise(resolve => setTimeout(() => resolve(), ms));

const loop = async <I extends {}, R extends {}>(
  f: () => Promise<R>,
  max: number,
  stop: (r: R) => boolean,
  i: number
): Promise<R> => {
  const r = await f();
  if (stop(r) || i >= max) {
    return r;
  } else {
    await wait(10000);
    return loop(f, max, stop, i + 1);
  }
};
export async function checkSubscription(window: Window, maybeConfig: IConfig, strategy?:IFlow): Promise<string> {
  const config: Config = strategy ? {
    ...strategy.flowConfig
  } : {...getConfig(), ...(maybeConfig || {})}
  const { host } = config

  const rockmanId = window.pac_analytics.visitor.rockmanId
  const check : () => Promise<ICheckSubscriptionResult> = () => fetch(`https://${host}/tallyman/v1/?action=check-subscription&rockman_id=${rockmanId}`)
  .then(x => x.json())

  const checkResult: ICheckSubscriptionResult = await loop(
    check,
    60,
    r => r.success,
    0
  )

  if(true === checkResult.success) {
    window.location.href = checkResult.product_url;
    return checkResult.product_url || null
  } else {
    const error = new Error(`${"SubscriptionFailed"}:\n${checkResult.message}`)
    error['type'] = "SubscriptionFailed";
    throw error
  }
}

// {"success":true,"uid":"*5QADM","shortcode":"1901","keyword":"TIPS"}%  
// curl "http://w1.mozzi.com/spa-api/?country=gh&device=smart&slug=18plus-oc2sms&action=oc2sms"