import * as React from "react";
import submitMSISDN, { IConfig, IKeywordShortcode, checkSubscription } from "./main";
import * as RDS from "../../common-types/RemoteDataState";
import { ITracker } from "../../pacman/record";
import { MOLink, formatSMSLink } from "../common-components/MOLink";
import { IFlow } from "../super-strategy/HOC";

export type MSISDNEntryFailure = {
  errorType: MSISDNEntryErrorTypes;
  error: any;
};
export type MSISDNEntrySuccess = IKeywordShortcode & { MOLink: ACompType }
type ACompType =
  React.ComponentType<{ children: React.ReactNode } & React.HTMLAttributes<HTMLAnchorElement> & { keyword?: string, shortcode?: string }>

export type State =
  | {
    type: "MSISDNEntry";
    result: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess & {MOLink: ACompType}>;
  }
  | {
    type: 'Completed';
    result: void;
  };

export type MSISDNEntryErrorTypes =
  | "AlreadySubscribed"
  | "UnknownError"
  | "InvalidMSISDN";


const mkMOLink = (tracker, keywordAndShortcode) => ({ children, keyword, shortcode, ...props }) => {
  const keywordAndShortcode1 = {
    keyword: keyword || keywordAndShortcode.keyword,
    shortcode: shortcode || keywordAndShortcode.shortcode
  }
  return <MOLink onClick={() => {
    if(!tracker) {
      console.warn("tracker parameter is null in mkMOLink")
    } else {
      tracker.advancedInFlow("click2sms", "click", keywordAndShortcode1)
    }
  }}
    keywordAndShortcode={keywordAndShortcode1} {...props}
  >{children}</MOLink>
}

export type PINEntryErrorTypes = "UnknownError" | "TooEarly" | "InvalidPIN";

export const initialState: State = { type: "MSISDNEntry", result: RDS.NothingYet<MSISDNEntryFailure, MSISDNEntrySuccess>() }
export const mockedMSISDNEntrySuccess: State = { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess>({ keyword: 'TEST OK', shortcode: '666', MOLink: mkMOLink(null, { keyword: 'TEST OK', shortcode: '666'}) }) }
export const mockedMSISDNEntryFailure: State =  {type: "MSISDNEntry", result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({ errorType: "InvalidMSISDN", error: "Invalid Mobile Number" })}
export const mockedCompletedState: State = { type: "Completed", result: void 6 }


export function matchMSISDNEntry<R>(m: RDS.IMatcher<MSISDNEntryFailure, MSISDNEntrySuccess & { MOLink: ACompType }, R>) { return RDS.match<MSISDNEntryFailure, MSISDNEntrySuccess & { MOLink: ACompType }, R>(m) }

export interface IActions {
  submitMSISDN: (window: Window, config: IConfig, msisdn: string, strategy?:IFlow, extraParams?: {[key: string]: string}) => void,
  backToStart: () => void
}

export type HOCProps = {
  currentState?: State;
  actions?: IActions;
  strategyConfig?:IFlow;
};

export function match<R>(
  { msisdnEntry, completed }:
    {
      msisdnEntry: (rds: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess>) => R
      , completed: (result: void) => R
    }): (state: State) => R {
  return state => {
    switch (state.type) {
      case 'MSISDNEntry':
        return msisdnEntry(state.result)
      case 'Completed':
        return completed(state.result)
    }
  }
}

export default <P extends HOCProps>(tracker: ITracker, Comp: React.ComponentType<P>, strategy?:IFlow) => (
  initialState: State
) =>{
  const HOC = class HOC extends React.PureComponent<P> {
    state: {
      currentState: State;
      actions: IActions;
    };
    constructor(props) {
      super(props);
      const self = this;
      this.state = {
        currentState: initialState,
        actions: {
          backToStart: () => {
            this.setState({
              currentState: initialState
            })
          },
          submitMSISDN: async (...args) => {
            this.setState({
              currentState: { type: "MSISDNEntry", result: RDS.Loading() } as State
            })
            const [, , msisdn] = args
            tracker.advancedInFlow('tallyman.v1-mo', 'msisdn-submitted', { msisdn })
            if(strategy){
              args[3] = strategy;
            }

            try {
              const keywordAndShortcode = await submitMSISDN(...args);
              tracker.advancedInFlow('tallyman.v1-mo', 'msisdn-submission-success', { msisdn })

              const moLink = mkMOLink(tracker, keywordAndShortcode)
              
              self.setState({
                currentState: { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess & { MOLink: ACompType }>({...keywordAndShortcode, MOLink: moLink}) } as State,
                actions: { ...self.state.actions }
              })


              try{
                checkSubscription(...args)
                .then((product_url)=>{

                  tracker.advancedInFlow('tallyman.v1-mo', 'sale', { msisdn });
                  setTimeout(() => {
                    window.location.href = product_url
                  }, 1000);
                })
              }catch(err){
                throw err;  
              }


            } catch (ex) {
              const errorType: MSISDNEntryErrorTypes =
                "SEAlreadySubscribed" === ex.type
                  ? "AlreadySubscribed"
                  : "SEInvalidMSISDN" == ex.type
                    ? "InvalidMSISDN"
                    : "UnknownError";
              self.setState({
                currentState: {
                  type: "MSISDNEntry",
                  result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({
                    errorType: errorType,
                    error: ex
                  })
                } as State
              });
              tracker.recedeInFlow('tallyman.v1-mo', 'msisdn-submission-failure', { msisdn, errorType: errorType || 'UnknownError' })
            }
          }
        }
      };
    }

    render() {
      return (
        <Comp {...this.props} currentState={this.state.currentState} actions={this.state.actions} strategyConfig={strategy ? strategy : undefined}/>
      );
    }
  };
  HOC.prototype['_client'] = "tallyman.v1-mo";

  return HOC;
}
