import * as React from "react";
import submitMSISDN, { IConfig } from "./main";
import * as RDS from "../../common-types/RemoteDataState";
import { ITracker } from "../../pacman/record";
import { IFlow } from "../super-strategy/HOC";

export const Link: React.ComponentType<{ href: string, children: React.ReactNode } & React.HTMLAttributes<HTMLAnchorElement>> = ({ href, children, ...props }) =>
  <a href={href} {...props}>{children}</a>

export type MSISDNEntryFailure = {
  errorType: MSISDNEntryErrorTypes;
  error: any;
};
export type MSISDNEntrySuccess = { Link: ACompType, href: string }
type ACompType =
  React.ComponentType<{ children: React.ReactNode } & React.HTMLAttributes<HTMLAnchorElement>>

export type State =
  | {
    type: "MSISDNEntry";
    result: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }>;
  }
  | {
    type: 'Completed';
    result: void;
  };

export type MSISDNEntryErrorTypes =
  | "AlreadySubscribed"
  | "UnknownError"
  | "InvalidMSISDN";


const mkLink = (tracker: ITracker, href: string) => ({ children, keyword, shortcode, ...props }) => {
  return <Link onClick={() => {
    if (!tracker) {
      console.warn("tracker parameter is null in mkLink")
    } else {
      tracker.advancedInFlow("click2sms", "click", href)
    }
  }}
    href={href} {...props}
  >{children}</Link>
}

export type PINEntryErrorTypes = "UnknownError" | "TooEarly" | "InvalidPIN";

export const initialState: State = { type: "MSISDNEntry", result: RDS.NothingYet<MSISDNEntryFailure, MSISDNEntrySuccess>() }
export const mockedMSISDNEntrySuccess: State = { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess>({ href: 'https://www.google.com', Link: mkLink(null, 'https://www.google.com') }) }
export const mockedMSISDNEntryFailure: State = { type: "MSISDNEntry", result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({ errorType: "InvalidMSISDN", error: "Invalid Mobile Number" }) }
export const mockedCompletedState: State = { type: "Completed", result: void 6 }


export function matchMSISDNEntry<R>(m: RDS.IMatcher<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }, R>) { return RDS.match<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }, R>(m) }

export interface IActions {
  submitMSISDN: (window: Window, config: IConfig, internationalMsisdn: string, bupperMsisdn: string, strategy?:IFlow, extraParams?: {[key: string]: string;}) => void,
  backToStart: () => void
}

export type HOCProps = {
  currentState: State;
  actions: IActions;
  strategyConfig?:IFlow;
};

export function match<R>(
  { msisdnEntry, completed }:
    {
      msisdnEntry: (rds: RDS.RemoteDataState<MSISDNEntryFailure, MSISDNEntrySuccess>) => R
      , completed: (result: void) => R
    }): (state: State) => R {
  return state => {
    switch (state.type) {
      case 'MSISDNEntry':
        return msisdnEntry(state.result)
      case 'Completed':
        return completed(state.result)
    }
  }
}

export default <P extends HOCProps>(tracker: ITracker, Comp: React.ComponentType<P>, strategy?:IFlow) => (
  initialState: State
) => {
  const HOC = class HOC extends React.PureComponent<P> {
    state: {
      currentState: State;
      actions: IActions;
    };
    constructor(props) {
      super(props);
      const self = this;
      this.state = {
        currentState: initialState,
        actions: {
          backToStart: () => {
            this.setState({
              currentState: initialState
            })
          },
          submitMSISDN: async (...args) => {
            this.setState({
              currentState: { type: "MSISDNEntry", result: RDS.Loading() } as State
            })
            const [, , msisdn] = args
            tracker.advancedInFlow('tallyman.v1-mo-redir', 'msisdn-submitted', { msisdn })
            if(strategy){
              args[4] = strategy;
              if(strategy.operator){
                args[5] = {operator:strategy.operator};
              }
            }

            try {
              const href = await submitMSISDN(...args);
              tracker.advancedInFlow('tallyman.v1-mo-redir', 'msisdn-submission-success', { msisdn })

              const moLink = mkLink(tracker, href)

              self.setState({
                currentState: { type: "MSISDNEntry", result: RDS.Success<MSISDNEntryFailure, MSISDNEntrySuccess & { Link: ACompType }>({ href, Link: moLink }) } as State,
                actions: { ...self.state.actions }
              })
            } catch (ex) {
              const errorType: MSISDNEntryErrorTypes =
                "SEAlreadySubscribed" === ex.type
                  ? "AlreadySubscribed"
                  : "SEInvalidMSISDN" == ex.type
                    ? "InvalidMSISDN"
                    : "UnknownError";
              self.setState({
                currentState: {
                  type: "MSISDNEntry",
                  result: RDS.Failure<MSISDNEntryFailure, MSISDNEntrySuccess>({
                    errorType: errorType,
                    error: ex
                  })
                } as State
              });
              tracker.recedeInFlow('tallyman.v1-mo-redir', 'msisdn-submission-failure', { msisdn, errorType: errorType || 'UnknownError' })
            }
          }
        }
      };
    }

    render() {
      return (
        <Comp {...this.props} currentState={this.state.currentState} actions={this.state.actions} strategyConfig={strategy ? strategy : undefined}/>
      );
    }
  };
  HOC.prototype['_client'] = "tallyman.v1-moredir";

  return HOC;
}