import * as React from "react"
import { ITracker } from "../../pacman/record";
import { IFlow } from "../super-strategy/HOC";
import * as RDS from "../../common-types/RemoteDataState";

import getRedirectUrl ,{ IConfig, ILinkExtraParams } from "./main";

export type IdentifyUserErrorTypes = "NotOn3g" | "UnknownError";
export type SubscribeUserErrorTypes = "SubscriptionError" | "UnknownError";

export type IdentifyFailure = {
  errorType: IdentifyUserErrorTypes;
  error: Error;
};


export type IdentifySuccess = {
  redirect_url?:string;
};

export type SubscribeSuccess = {
  product_url:string;
};

export type SubscribeFailure = {
  errorType: SubscribeUserErrorTypes;
  error: Error;
};


export type IState =
| {
    type: "IDENTIFY";
    result: RDS.RemoteDataState<IdentifyFailure, IdentifySuccess>;
  }
| {
    type: "SUBSCRIBE";
    result: RDS.RemoteDataState<SubscribeFailure, SubscribeSuccess>;
  }


export const initialState : IState = { type: "IDENTIFY", result: RDS.NothingYet<IdentifyFailure, IdentifySuccess>()}
export const mockedSubscribeState : IState = { type: "SUBSCRIBE", result: RDS.NothingYet<SubscribeFailure, SubscribeSuccess>() }
export const mockedCompletedState : IState = { type: "SUBSCRIBE", result: RDS.Success<SubscribeFailure, SubscribeSuccess>({product_url: 'https://www.yahoo.com/'}) }
type IActions = {
  identifyAction:()=>any;
  subscribeAction?:()=>any;
}

type ACompType = 
    React.ComponentType<{children: React.ReactNode} & React.HTMLAttributes<HTMLAnchorElement>>


export type HOCProps= {
  currentState: IState;
  actions: IActions;
  ProductLink: ACompType;
  RedirectLink: ACompType;
  strategyConfig?:IFlow;
  trackAdvanceInflow?: () => void;
};

export function match<R>(
  { identify, subscribe }: 
  {   identify: (rds: RDS.RemoteDataState<IdentifyFailure, IdentifySuccess>) => R
    , subscribe: (rds: RDS.RemoteDataState<SubscribeFailure, SubscribeSuccess>) => R 
  }): (state: IState) => R {
  return state => {
    switch (state.type) {
      case 'IDENTIFY':
        return identify(state.result)
      case 'SUBSCRIBE':
        return subscribe(state.result)
    }
  }
}


export default  <P extends HOCProps>(tracker: ITracker, maybeConfig: IConfig, Comp: React.ComponentType<HOCProps>, extraParams?:ILinkExtraParams, strategy?:IFlow) => (initialState: IState, isMocked?:boolean) => {
  const HOC = class HOC extends React.PureComponent<P> {
    state:{
      currentState: IState;
      actions:IActions;
      product_url:string;
      redirect_url:string;
      is_clicked?:boolean;
    }
    constructor(props) {
      super(props);
      const self = this;
      this.state = {
        currentState: initialState,
        actions: {
          identifyAction: async ()=>{
            try{
              self.setState({
                currentState:{ type:"IDENTIFY", result: RDS.Loading()} as IState
              });
              const {subscription_url, redirect_url, action} = await getRedirectUrl(maybeConfig || {}, extraParams, strategy);

              if(redirect_url){
                tracker.customEvent('Flow', 'advance-auto', 'redirect-success', {redirect_url})
                self.setState({
                  currentState: {type: "IDENTIFY", result: RDS.Success({redirect_url})}  as IState,
                  redirect_url
                });
              }else{
                tracker.customEvent('Flow', 'advance-auto', 'identify-success', {subscription_url})

                self.setState({
                  currentState:{ type:"SUBSCRIBE", result: RDS.NothingYet()}  as IState,

                  actions: {
                    ...self.state.actions,
                    subscribeAction: async () => {
                      self.setState({
                        currentState: {type: "SUBSCRIBE", result: RDS.Loading()} as IState
                      });
                      try{
                        tracker.advancedInFlow('tallyman.v1-one-click', 'click-subscribe');
                        const productUrl = await action();
                        self.setState({
                          currentState: {type: "SUBSCRIBE", result: RDS.Success({productUrl})}  as IState,
                          productUrl
                        });
                        tracker.advancedInFlow('tallyman.v1-one-click', 'click-subscribe-success', {product_url:productUrl});
                      }catch(ex){
                        console.error(ex)
                        const errorType: SubscribeUserErrorTypes =
                          "SubscriptionError" === ex.type
                            ? "SubscriptionError"
                            : "UnknownError";

                        tracker.recedeInFlow('tallyman.v1-one-click', 'click-subscribe-failure', {error:ex.toString()});
                        self.setState({
                          currentState: {type: "SUBSCRIBE", result: RDS.Failure("IError")}  as IState
                        });

                        self.setState({
                          currentState:{
                            type:"SUBSCRIBE",
                            result: RDS.Failure<SubscribeFailure, SubscribeSuccess>({
                              errorType: errorType,
                              error: ex
                            })
                          }  as IState
                        });
                      }
                    }
                  }
                })
              }
            }catch(ex) {
              console.error(ex)
              const errorType: IdentifyUserErrorTypes =
                "NotOn3g" === ex.type
                  ? "NotOn3g"
                  : "UnknownError";
              
              tracker.customEvent('Flow', 'recede-auto', 'identify-failure', {error:ex.toString(), errorType})
              this.setState({
                currentState:{
                  type:"IDENTIFY",
                  result: RDS.Failure<IdentifyFailure, IdentifySuccess>({
                    errorType: errorType,
                    error: ex
                  })
                } as IState
              });
            }

          }
        },
        product_url:"",
        redirect_url:""
      }
    }

    async UNSAFE_componentWillMount() {
      {
        !isMocked &&
        await this.state.actions.identifyAction()
      }
    }

    componentDidUpdate(){
      if(this.state.is_clicked && this.state.redirect_url !== ""){
        window.location.href=this.state.redirect_url;
      }
    }
    render() {
      const self = this;
      const productLink = ({children, ...props}) => <a {...props} onClick={(...args) => {
        tracker.advancedInFlow('tallyman.v1-one-click', 'click-product-url', {product_url:self.state.product_url});
        !!props.onClick ? props.onClick(...args) : void 6;
      }} href={self.state.product_url}>{children}</a>

      const redirectLink = ({children, ...props}) => <a {...props} onClick={(...args) => {
        tracker.advancedInFlow('tallyman.v1-one-click', 'click-redirect-url', {redirect_url:self.state.redirect_url});
        this.setState({is_clicked:true})
        !!props.onClick ? props.onClick(...args) : void 6;
      }} href={self.state.redirect_url || null}>{children}</a>
      return (
        <Comp
          actions={self.state.actions}
          ProductLink={productLink}
          RedirectLink={redirectLink}
          currentState={self.state.currentState}
          strategyConfig={strategy ? strategy : undefined}
          {...this.props}
        />
      );
    }
  }

  HOC.prototype['_client'] = "tallyman.v1-one-click";
  return HOC;
}
