//import { getConfig, ConfigDetails} from "./config"
import { IFlow } from "../super-strategy/HOC";
export type IConfig = {
  offer: number,
  slug?: string;
  country?: string;
  host?:string;
  device?: "smart" | "feature";
  automaticallySubmitAllOperators?: boolean;
  operators?: string[];
}
  
export type IMSISDNSubmissionResult = {
  success: true
  rockman_id: string
  pin: string
  product_url?: string
} | {
  success: false
  rockman_id: string
  message: string
  product_url?: string
 }

export type IPINSubmissionResult = {
  success: true
  rockman_id: string
  product_url?: string
  async?: boolean;
} | {
  success: false
  rockman_id: string
  message: string
}

export type ICheckSubscriptionResult = {
  success: false;
  rockman_id: string;
  message: string;
} | {
  success: true;
  msisdn: string;
  product_url?: string;
  rockman_id: string;
}

export interface IHash {
  [key: string]: string
}

export type SingleMSISDNSubmissionResult = {
  type: "SingleMSISDNSubmissionResult",
  pin?: string,
  submitPIN: (pin: string) => Promise<string>
}

export type MultiOperatorMSISDNSubmissionResult = {
  type: "MultiOperatorMSISDNSubmissionResult",
  submitPIN: (pin: string) => Promise<string>
}



// const defaultConfig = (offer: number, country?: string, scenario?: string) : IConfig => {
//   return {...getConfig(country || process.env.country, scenario || process.env.scenario).config, offer}
// }

const wait = (ms: number) =>
  new Promise(resolve => setTimeout(() => resolve(), ms));

const loop = async <I extends {}, R extends {}>(
  f: () => Promise<R>,
  max: number,
  stop: (r: R) => boolean,
  i: number
): Promise<R> => {
  const r = await f();
  if (stop(r) || i >= max) {
    return r;
  } else {
    await wait(1000);
    return loop(f, max, stop, i + 1);
  }
};


const getConfig = ():IConfig=>{
  try {
    const configs:IConfig = require("../../../../../config.json");
    return configs;
  } catch (ex) {
    throw "config.json not found";
  }
};

export default async function submitMSISDN(window: Window, maybeConfig: IConfig, internationalMSISDN: string, strategy?:IFlow, extraParams?: IHash): Promise<SingleMSISDNSubmissionResult | MultiOperatorMSISDNSubmissionResult> {
 

  const config: IConfig = strategy ? {
    offer:window.pac_analytics.visitor.offer,
    ...strategy.flowConfig
  } : {offer:window.pac_analytics.visitor.offer, ...getConfig(), ...(maybeConfig||{})}
  if(true == config.automaticallySubmitAllOperators && !!config.operators && config.operators.length > 0) {
    return submitMSISDNForMultipleOperators(window, config, internationalMSISDN, config.operators, extraParams)
  } else {
    return submitMSISDNOnce(window, config, internationalMSISDN, extraParams)
  }
}

async function submitMSISDNOnce(window: Window, config: IConfig, internationalMSISDN: string, extraParams?: IHash): Promise<SingleMSISDNSubmissionResult> {
  if (!internationalMSISDN ) {
    const submissionError = new Error(`Error in submitMSISDN() trigger-pin action`)
    submissionError['type'] = "InvalidMSISDN";
    throw submissionError
  }

  const msisdn = internationalMSISDN.match(/\d+/gi).join('') // tallyman API expects international msisdn without any special character
  const { slug, country, device, offer, host } = config
  const bupperizeCountry = c => c == 'gb' ? 'uk' : c;
  const search = window.location.search.substr(1) || ''
  const extraParamsQs = !extraParams ? '' : '&' + Object.keys(extraParams).map(k => `${k}=${extraParams[k]}`).join('&')
  const rockmanId = window.pac_analytics.visitor.rockmanId
  const result: IMSISDNSubmissionResult = await fetch(`https://${host}/tallyman/v1/?action=trigger-pin&country=${bupperizeCountry(country)}&slug=${slug}&device=${device}&offerId=${offer}&msisdn=${msisdn}&rockman_id=${rockmanId}${extraParamsQs}&${search}`).then(x => x.json())
  if (false === result.success) {
    const error = new Error(`Error in submitMSISDN() trigger-pin action:\n${result.message}`)
    error['type'] = result.message == 'ALREADY SUBSCRIBED' ? "AlreadySubscribed" : "InvalidMSISDN";
    error['productUrl'] = result.product_url ? result.product_url : "";
    console.error(error)
    throw error
  } else {
    return {
      type: "SingleMSISDNSubmissionResult",
      pin: result.pin,
      submitPIN: async function submitPIN(pin: string): Promise<string> {
        if(!pin){
          const emptyPinError = new Error(`Error in submitMSISDN() verify-pin action:\n Pin Cannot Be Empty`);
          emptyPinError['type'] = "InvalidPIN";
          throw emptyPinError
        }
        const pinResult: IPINSubmissionResult = await fetch(`https://${host}/tallyman/v1/?action=verify-pin&country=${bupperizeCountry(country)}&slug=${slug}&device=${device}&offerId=${offer}&rockman_id=${rockmanId}&pin=${pin}${extraParamsQs}`).then(x => x.json())
        if (false === pinResult.success) {
          const pinError = new Error(`Error in submitMSISDN() verify-pin action:\n${pinResult.message}`)
          pinError['type'] = "InvalidPIN";
          throw pinError
        } else {

          if(true === pinResult.async) {
            const check : () => Promise<ICheckSubscriptionResult> = () => fetch(`https://${host}/tallyman/v1/?action=check-subscription&country=${bupperizeCountry(country)}&slug=${slug}&device=${device}&offerId=${offer}&rockman_id=${rockmanId}&pin=${pin}${extraParamsQs}`)
              .then(x => x.json())

            const checkResult: ICheckSubscriptionResult = await loop(
              check,
              30,
              r => r.success,
              0
            )

            if(true === checkResult.success) {
              return pinResult.product_url || checkResult.product_url || null
            } else {
              const pinError = new Error(`Error in submitMSISDN() verify-pin action:\n${checkResult.message}`)
              pinError['type'] = "InvalidPIN";
              throw pinError
            }
          } else {
            return pinResult.product_url || null
          }
        }
      }
    }
  }
}

type Either<L, R> = { type: 'left', value: L } | {type: 'right', value: R}
function right<L, R>(r: R) {
  return {type: 'right', value: r} as Either<L, R>
}
function left<L, R>(l: L) {
  return { type: 'left', value: l } as Either<L, R>
}
function foldEither<L, R, T>(left: (l:L) => T, right: (r:R) => T, either: Either<L, R>) : T {
  return either.type == 'left' 
    ? left(either.value)
    : right(either.value)
}

async function submitMSISDNForMultipleOperators(window: Window, config: IConfig, internationalMSISDN: string, operators: string[], extraParams?: IHash): Promise<MultiOperatorMSISDNSubmissionResult> {
  const results = await Promise.all(operators.map(async o => {
    try {
      const singleResult = await submitMSISDNOnce(window, config, internationalMSISDN, { ...extraParams, operator: o })
      return right({...singleResult, operator: o})
    } catch (error) {
      return left(error)
    }
  }
  ))

  //@ts-ignore
  const alreadySubscribed = results.find(r => r.type == 'left' && r.value.type == 'AlreadySubscribed')

  if (results.every(r => r.type == 'left')) {
    // no submission worked
    throw results[0].value
  } else if (!!alreadySubscribed) {
    throw alreadySubscribed.value
  } else {
    // const pins = results.filter(r => r.type == 'right').map(r => r.value as SingleMSISDNSubmissionResult).map(r => r.pin)
    return {
      type: "MultiOperatorMSISDNSubmissionResult",
      submitPIN: async function submitPIN(pin: string): Promise<string> {
        const pinResults = await Promise.all(
          results.filter(r => r.type == 'right').map(r => r.value as SingleMSISDNSubmissionResult)
            .map(async ({submitPIN}) => {
              try {
                const res = await submitPIN(pin)
                return right(res) as Either<{ error: Error }, string>
              } catch (error) {
                return left(error) as Either<{ error: Error }, string>
              }
            })
        )
        if (pinResults.every(p => p.type == 'left')) {
          // no pin submission worked
          throw results[0].value
        } else {
          return pinResults.find(p => p.type == 'right').value as string
        }
      }
    }
  }
}
