import {IFlowEvent, IPacman} from "./types";

import { map } from "rxjs/operators/map";
import { bufferWhen } from "rxjs/operators/bufferWhen";
import { filter } from "rxjs/operators/filter";
import { debounceTime } from "rxjs/operators/debounceTime";
import { first } from "rxjs/operators/first";

import { merge } from "rxjs/observable/merge";
import { fromEvent } from "rxjs/observable/fromEvent";
import { from } from "rxjs/observable/from";
import { interval } from "rxjs/observable/interval";
import { empty } from "rxjs/observable/empty";
import { zip } from "rxjs/observable/zip";
import { ReplaySubject } from 'rxjs/ReplaySubject';

import IVisitor from "../common-types/IVisitor";

// (Homam:) I disabled canvasHash for now, we can get it back by un-commenting the lines prefixed with //CanvasHash:
//CanvasHash: import md5 from './md5';
//CanvasHash: import canvasHash from './canvasHash';

export class Pacman implements IPacman {

  private flow_event;
  private custom;
  private flowEventNumber: number;
  private batch: number;

  public rockmanId:string;
  public impressionNumber:number
  public startTime: number;

  readonly serverUrl = process.env.pacman_server // check webpack.config.common.js

  private d: Document;
  private w: Window;

  constructor(w: Window, visitor: IVisitor, startTime: number) {
    this.flowEventNumber = 1;
    this.batch = 0;

    this.d = w.document;
    this.w = w;

    this.rockmanId = visitor.rockmanId;
    this.impressionNumber = visitor.impressionNumber
    this.startTime = startTime;
  }

  initInteractions() {

    let thisLocal = this;
    let w = this.w;
    let d = this.d;

    let documentProp = new DocumentProp(w);
    
    var visibilityProperty = "hidden",
        visibilityChangeEvent: string;
    if (visibilityProperty in d) visibilityChangeEvent = "visibilitychange";
    else if ((visibilityProperty = "mozHidden") in d)
      visibilityChangeEvent = "mozvisibilitychange";
    else if ((visibilityProperty = "webkitHidden") in d)
      visibilityChangeEvent = "webkitvisibilitychange";
    else if ((visibilityProperty = "msHidden") in d)
      visibilityChangeEvent = "msvisibilitychange";

    var visibilities = visibilityChangeEvent
        ? fromEvent(d, visibilityChangeEvent).pipe(
            map(function() {
              return {
                t: d[visibilityProperty] ? "hidden" : "visible",
                s: thisLocal.getRelativeTime()
              };
            })
        )
        : empty();

    var unload = fromEvent(w, "beforeunload");
    var blur = fromEvent(w, "blur").pipe(
        map(function() {
          return {
            t: "blur",
            s: thisLocal.getRelativeTime(),
            a: {
              e: "window"
            }
          };
        })
    );
    var focus = fromEvent(w, "focus").pipe(
        map(function() {
          return {
            t: "focus",
            s: thisLocal.getRelativeTime(),
            a: {
              e: "window"
            }
          };
        })
    );
    var focusin = fromEvent(d, "focusin").pipe(
        map(function(event) {
          return {
            t: "focus",
            s: thisLocal.getRelativeTime(),
            a: {
              e: documentProp.getEventTarget(event).id
            }
          };
        })
    );
    var focusout = fromEvent(d, "focusout").pipe(
        map(function(event) {
          return {
            t: "blur",
            s: thisLocal.getRelativeTime(),
            a: {
              e: documentProp.getEventTarget(event).id
            }
          };
        })
    );
    var submit = fromEvent(d, "submit").pipe(
        map(function(event) {
          return {
            t: "submit",
            s: thisLocal.getRelativeTime(),
            a: {
              e: documentProp.getEventTarget(event).id
            }
          };
        })
    );

    var firstChange = fromEvent(d, "change").pipe(
        map(function(event) {
          return {
            t: "fchange",
            s: thisLocal.getRelativeTime(),
            a: {
              e: documentProp.getEventTarget(event).id
            }
          };
        }),
        first()
    );

    var firstTouch = fromEvent(d, "touchstart").pipe(
        map(function() {
          return {
            t: "ftouch",
            s: thisLocal.getRelativeTime()
          };
        }),
        first()
    );

    var firstMotion = fromEvent(w, "devicemotion").pipe(
        map(function() {
          return {
            t: "fmotion",
            s: thisLocal.getRelativeTime()
          };
        }),
        first()
    );

    var firstAccelerationIncludingGravity;
    var firstMotion = fromEvent(w, 'devicemotion').pipe(
        map(function (event) {
          return event.accelerationIncludingGravity
        }),
        first(function (event) {
          if (firstAccelerationIncludingGravity) {
            return !!
                (Math.abs(firstAccelerationIncludingGravity.x - event.x) > 0.2 ||
                    Math.abs(firstAccelerationIncludingGravity.y - event.y) > 0.2 ||
                    Math.abs(firstAccelerationIncludingGravity.z - event.z) > 0.2)
          } else {
            firstAccelerationIncludingGravity = event
            return false
          }
        })
    ).pipe(
        map(function() {
          return {
            "t": 'fmotion',
            "s": thisLocal.getRelativeTime()
          }
        })
    );


    var full_load = fromEvent(w, "load").pipe(
        map(function() {
          return {
            t: "full_load",
            s: thisLocal.getRelativeTime(),
            a: {
              p: documentProp.getViewportSize()
            }
          };
        })
    );

    var html_load;
    if (this.d.readyState === "complete" || this.d.readyState === "interactive") {
      html_load = from([
        {
          t: "html_load",
          s: thisLocal.getRelativeTime(),
          a: {
            p: documentProp.getViewportSize()
          }
        }
      ]);
    } else {
      html_load = fromEvent(d, "DOMContentLoaded").pipe(
          map(function() {
            return {
              t: "html_load",
              s: thisLocal.getRelativeTime(),
              a: {
                p: documentProp.getViewportSize()
              }
            };
          })
      );
    }

    this.flow_event = new ReplaySubject();

    var touchstart = fromEvent<TouchEvent>(d, "touchstart");
    var touchend = fromEvent<TouchEvent>(d, "touchend");

    this.custom = new ReplaySubject();

    /*CanvasHash:
    var fp_canvas = Observable.create(function (observer) {
      if (w.requestIdleCallback) {
        w.requestIdleCallback(function () {
          observer.next({
            "t": 'fp_canvas',
            "s": thisLocal.getRelativeTime(),
            "a": {
              "e": md5(canvasHash())
            }
          })
        })
      } else {
        setTimeout(function () {
          observer.next({
            "t": 'fp_canvas',
            "s": thisLocal.getRelativeTime(),
            "a": {
              "e": md5(canvasHash())
            }
          })
        }, 700)

      }
    });
    */

    var clicks = zip(touchstart, touchend).pipe(
        filter(function(x) {
          return (
              x[1].timeStamp - x[0].timeStamp < 350 &&
              x[0].touches.length == 1 &&
              x[1].touches.length == 0
          );
        }),
        map(function(event) {
          return {
            t: "click",
            s: thisLocal.getRelativeTime(),
            a: {
              e: documentProp.getEventTarget(event[0].touches[0]).id,
              p: documentProp.getEventPos(event[0].touches[0])
            }
          };
        })
    );

    var scrolls = fromEvent(w, "scroll").pipe(
        debounceTime(500),
        map(function(event) {
          return {
            t: event.type,
            s: thisLocal.getRelativeTime(),
            a: {
              p: documentProp.getDocumentScroll()
            }
          };
        })
    );

    var resize = fromEvent(w, "resize").pipe(
        debounceTime(500),
        map(function(event) {
          return {
            t: event.type,
            s: thisLocal.getRelativeTime(),
            a: {
              p: documentProp.getViewportSize()
            }
          };
        })
    );

    var stream = merge(
        clicks,
        scrolls,
        blur,
        focus,
        focusin,
        focusout,
        visibilities,
        firstTouch,
        firstChange,
        firstMotion,
        submit,
        resize,
        full_load,
        html_load,
        //CanvasHash: fp_canvas,
        this.flow_event,
        this.custom
    ).pipe(
        bufferWhen(function() {
          return merge(interval(7000), unload, blur);
        }),
        filter(function(x) {
          return x.length > 0;
        })
    );

    stream.subscribe(function(data) {
      var xhr = new XMLHttpRequest();
      xhr.open("POST", thisLocal.serverUrl, true);
      xhr.send(
          JSON.stringify({
            r: thisLocal.rockmanId,
            m: thisLocal.impressionNumber,
            b: thisLocal.batch++,
            d: data
          })
      );
    });
  }

  sendPageview() {

    let thisLocal = this;
    let w = this.w;
    let d = this.d;
    let n = this.w.navigator;

    var vp = "hidden";
    vp in d ||
    (vp = "mozHidden") in d ||
    (vp = "webkitHidden") in d ||
    (vp = "msHidden") in d;
    var r = function(f) {
          try {
            var o = f.apply(this, arguments);
            return typeof o == typeof true ? +o : o;
          } catch (e) {
            return null;
          }
        },
        ua = n.userAgent.toLowerCase(),
        re =
            "CSS1Compat" == document.compatMode
                ? document.documentElement
                : document.body || document.documentElement,
        pv = {
          t: "pageview",
          s: 0,
          tz: -new Date().getTimezoneOffset(),
          n: r(function() {
            return Notification.permission;
          }),
          f: r(function() {
            return w.top != w.self;
          }),
          nt: r(function() {
            return n.connection.type || null;
          }),
          ns: r(function() {
            return n.connection.effectiveType || null;
          }),
          l: r(function() {
            var a = w.navigator
                ? n.language ||
                n.userLanguage ||
                n.browserLanguage ||
                n.systemLanguage
                : null;
            return "[object String]" == Object.prototype.toString.call(a)
                ? a.toLowerCase()
                : null;
          }),
          lz: r(function() {
            return n.languages.join(" ").toLowerCase();
          }),
          vw: r(function() {
            return re.clientWidth || w.innerWidth;
          }),
          vh: r(function() {
            return re.clientHeight || w.innerHeight;
          }),
          sw: r(function() {
            return w.screen.width;
          }),
          sh: r(function() {
            return w.screen.height;
          }),
          sd: w.screen.colorDepth || null,
          sr: w.devicePixelRatio || null,
          np: r(function() {
            return n.platform || null;
          }),
          dnt: r(function() {
            if (typeof n.doNotTrack !== "undefined" && n.doNotTrack !== null) {
              if (n.doNotTrack == "yes") {
                return true;
              } else if (n.doNotTrack == "no") {
                return false;
              }
              return !!n.doNotTrack;
            } else if (
                typeof w.doNotTrack !== "undefined" &&
                w.doNotTrack !== null
            ) {
              return !!w.doNotTrack;
            }
            return null;
          }),
          c: r(function() {
            return n.cookieEnabled;
          }),
          ss: r(function() {
            return !!w.sessionStorage;
          }),
          wd: r(function() {
            return !!n.webdriver;
          }),
          ls: r(function() {
            return !!w.localStorage;
          }),
          ib: r(function() {
            return !!w.indexedDB;
          }),
          od: r(function() {
            return !!w.openDatabase;
          }),
          hf: r(function() {
            return !!d.hasFocus();
          }),
          iv: r(function() {
            return !!!d[vp];
          }),
          ll: r(function() {
            return n.languages[0].substr(0, 2) !== n.language.substr(0, 2);
          }),
          lr: r(function() {
            return (
                w.screen.width < w.screen.availWidth ||
                w.screen.height < w.screen.availHeight
            );
          }),
          lo: r(function() {
            var p = n.platform.toLowerCase();
            if (ua.indexOf("windows phone") >= 0 || ua.indexOf("win") >= 0)
              return null;
            if (ua.indexOf("android") >= 0)
              return (
                  typeof n.plugins === "undefined" ||
                  p.indexOf("win") >= 0 ||
                  p.indexOf("mac") >= 0 ||
                  p.indexOf("ipad") >= 0 ||
                  p.indexOf("ipod") >= 0 ||
                  p.indexOf("iphone") >= 0
              );
            if (ua.indexOf("linux") >= 0) return null;
            if (ua.indexOf("iphone") >= 0 || ua.indexOf("ipad") >= 0)
              return (
                  typeof n.plugins === "undefined" ||
                  p.indexOf("win") >= 0 ||
                  p.indexOf("linux") >= 0 ||
                  p.indexOf("android") >= 0 ||
                  p.indexOf("pike") >= 0
              );
            return null;
          }),
          lb: r(function() {
            var ps = n.productSub,
                t = eval.toString().length;
            if (ua.indexOf("firefox") >= 0) return t !== 37;
            if (ua.indexOf("opera") >= 0 || ua.indexOf("opr") >= 0)
              return ps !== "20030107" || t !== 33;
            if (ua.indexOf("chrome") >= 0) return ps !== "20030107" || t !== 33;
            if (ua.indexOf("safari") >= 0) return ps !== "20030107" || t !== 37;
            if (ua.indexOf("trident") >= 0) return t !== 39;
            return null;
          })
        };

    var xhr = new XMLHttpRequest();
    xhr.open("POST", thisLocal.serverUrl, true);
    xhr.send(
        JSON.stringify({
          r: thisLocal.rockmanId,
          m: thisLocal.impressionNumber,
          b: thisLocal.batch++,
          d: [{ t: "pageview", s: 0, a: pv }]
        })
    );
  }

  sendFlowEvent(flowEvent: IFlowEvent) {
    this.flow_event.next({
      t: 'flow_event',
      s: this.getRelativeTime(),
      a: {
        number: this.flowEventNumber++,
        ...flowEvent
      }
    })
  }

  sendCustomEvent(typeName: string, eventName: string) {
    this.custom.next({
      t: typeName,
      s: this.getRelativeTime(),
      a: {
        e: eventName
      }
    })
  };

  getRelativeTime() {
    return precisionRound((this.startTime ? new Date().valueOf() - this.startTime : 0) / 1000, 1);
  }
}

function precisionRound(number, precision) {
  var factor = Math.pow(10, precision);
  return Math.round(number * factor) / factor;
};

class DocumentProp {

  private d: Document;
  private w: Window;
  
  constructor(w: Window) {
    this.d = w.document;
    this.w = w;
  }

  getRootElement() {
    var element = this.d.documentElement;
    return "CSS1Compat" == this.d.compatMode ? element : this.d.body || element;
  }

  getViewportSize() {
    var rootElement = this.getRootElement();
    return [
      rootElement.clientWidth || this.w.innerWidth,
      rootElement.clientHeight || this.w.innerHeight
    ];
  }

  getDocumentScroll() {
    return [
      this.w.pageXOffset ||
      (this.d.documentElement && this.d.documentElement.scrollLeft) ||
      (this.d.body && this.d.body.scrollLeft) ||
      0,
      this.w.pageYOffset ||
      (this.d.documentElement && this.d.documentElement.scrollTop) ||
      (this.d.body && this.d.body.scrollTop) ||
      0
    ];
  }

  getEventPos(event) {
    var b = this.getRootElement(),
        c = this.getDocumentScroll();
    return [
      event.pageX || event.clientX + c[0] - (b.clientLeft || 0) || 0,
      event.pageY || event.clientY + c[1] - (b.clientTop || 0) || 0
    ];
  }

  getEventTarget(event) {
    var b = null;
    try {
      (b = event.target || event.srcElement) &&
      !b.ownerDocument &&
      b.documentElement &&
      (b = b.documentElement);
    } catch (c) {}
    return b;
  }
};