import mkSendBeacon from "./sendBeacon";
import queryString from "./queryString";
import {Pacman} from "./main";
import {IFlowEvent} from "./types";
import uuid from "uuid/v1";

export { queryString };

export function recordImpression(
  window: Window,
  url: string,
  country: string,
  page: string
) {
  if (!window.pac_analytics) {

    const rockmanId = uuid().replace(/-/g, "");

    const campaignId =
        parseInt(queryString(window.location.search, "cid")) || 2;

    window.pac_analytics = {
      visitor: {
        rockmanId,
        impressionNumber: 1, //TODO: get impressionNumber from localStorage
        country,
        page,
        xaid: queryString(window.location.search, "xaid"),
        cid: campaignId,
        offer: parseInt(queryString(window.location.search, "offer")) || 1 // default to 1 offer id
      },
      startTime: new Date().valueOf(),
      queryString: key => queryString(window.location.search, key),
      url,
      userId: rockmanId,
      encCampaignId: queryString(window.location.search, "xcid"),
      originalUrl: document.location.href,
    };

    const sendBeacon = mkSendBeacon(window);
    sendBeacon(`${url}/impression/${window.pac_analytics.encCampaignId}`, {
      rockmanId: window.pac_analytics.visitor.rockmanId,
      userId: window.pac_analytics.userId,
      page: window.pac_analytics.visitor.page,
      originalUrl: window.pac_analytics.originalUrl
    });
  } else {
    console.info("No need to record an impression from client");
    window.pac_analytics.queryString = key =>
      queryString(window.location.search, key);
  }

  // expect that window.pac_analytics[.visitor] is ready
  window.pacman = new Pacman(window, window.pac_analytics.visitor, window.pac_analytics.startTime);
  window.pacman.initInteractions();
  window.pacman.sendPageview();

  if (!window.dataLayer) {
    window.dataLayer = window.dataLayer || [];
  }

  window.dataLayer.push({ affiliate_id: window.pac_analytics.visitor.xaid });

  return window.pac_analytics;
}

export function recordEvent(
  window: Window,
  url: string,
  view: string,
  flowEvent: IFlowEvent
) {
  const sendBeacon = mkSendBeacon(window);
  if (!window.pac_analytics) {
    throw Error(
      "Please call recordImpression() first. 'pac_analytics' was not found."
    );
  }
  const eventDetail = {
    rockmanId: window.pac_analytics.visitor.rockmanId,
    relt: new Date().valueOf() - window.pac_analytics.startTime,
    view,
    ...flowEvent
  }
  sendBeacon(`${url}/event`, eventDetail);
  try{
    window.pacman.sendFlowEvent(flowEvent);
  }catch(err){
    console.warn(err)
  }
  
  try {
    const ev = new CustomEvent('tracking-event', {detail: eventDetail})
    document.body.dispatchEvent(ev)
  } catch(ex) {
    console.warn(ex)
  }
}

export type ITracker = {
  viewChanged: (view: string) => void;
  msisdnSubmitted: (msisdn: string) => void;
  advancedInFlow: (flow: string, action: string, args?: any) => void;
  advancedInPreFlow: (label: string, args?: any) => void;
  recedeInFlow: (flow: string, newState: string, args?: any) => void;
  customEvent:(category: string, action: string, label: string, arg?: any)=> void;
};

export default (window: Window, country: string, page: string): ITracker => {
  if (!window) {
    return {
      viewChanged: () => void 6,
      msisdnSubmitted: () => void 6,
      advancedInFlow: () => void 6,
      advancedInPreFlow: () => void 6,
      recedeInFlow: () => void 6,
      customEvent: () => void 6
    };
  } else {
    var current_view = "Init";

    const url = !!window.pac_analytics
      ? window.pac_analytics.url
      : queryString(window.location.search, "pacman-server") || "/analytics";

    // recordImpression manipulates window object by adding pac_analytics
    const pac_analytics = recordImpression(window, url, country, page);

    window.dataLayer.push({
      xaid: pac_analytics.visitor.xaid,
      country,
      page,
      offer: pac_analytics.visitor.offer
    });

    return {
      viewChanged: (view: string) => {
        if (view != current_view) {
          recordEvent(window, url, view, {
            category: "UI",
            action: "ViewChanged"
          });
          current_view = view;
        }
      },
      msisdnSubmitted: (msisdn: string) => {
        recordEvent(window, url, current_view, {
          category: "UX",
          action: "MSISDN-Submitted",
          args: { msisdn }
        });
      },
      advancedInPreFlow: (label: string, args?: any) => {
        const gaEvent = { category: "Pre-Flow", action: "advance", label };
        recordEvent(window, url, current_view, {
          ...gaEvent,
          args
        });
        window.dataLayer.push({ ...gaEvent, event: "gaEvent" });
        // how to trigger a custom event on window object:
        // const event = new CustomEvent('gaEvent', { detail: {category: 'Pre-Flow'} });
        // window.dispatchEvent(event)
      },
      recedeInFlow: (flow: string, reason: string, args?: any) => {
        const gaEvent = {
          category: "Flow",
          action: "recede",
          label: `${reason}`
        };
        recordEvent(window, url, current_view, {
          ...gaEvent,
          args
        });
        window.dataLayer.push({ ...gaEvent, event: "gaEvent" });
      },
      advancedInFlow: (flow: string, action: string, args?: any) => {
        const gaEvent = {
          category: "Flow",
          action: `advance`,
          label: `${action}`
        };
        recordEvent(window, url, current_view, {
          ...gaEvent,
          args
        });
        window.dataLayer.push({ ...gaEvent, event: "gaEvent" });
      },
      //Flow,   advance-auto, label : e.g indentify-success | indentify-failure, args : error | data | {}
      customEvent: (category: string, action: string, label: string, args?: any) => {
        const gaEvent = {
          category,
          action,
          label
        };
        recordEvent(window, url, current_view, {
          ...gaEvent,
          args
        });
        window.dataLayer.push({ ...gaEvent, event: "gaEvent" });
      }
    };
  }
};
